/*************************************************************************

	Project : ZipMe
	Author  : R'alf <mailto:raphael.moll@capway.com>

	Version : 0.7.4
	Date    : 01 Septembre 1997

	Format  : tabs == 2	

	Description :
	- The idea is to zip files, cope with it !
	- If only ONE folder or ONE file is selected, the file or the folder name
	  is used to make a "foo.zip" file.
	- If more than one item is selected, the whole stuff is packed under "archive.zip"
	  file name, with an incresing number if entry already exists.

	Version 0.7.4 :
	- new main() installer (moves the executable in user addon dir)

	Version 0.7.3 :
	- inserted changes proposed by Peter Hinely <phinely@hawaii.edu> [14/7/98]
	  he is modifying this to handle .boz archives (bzip over zip :-D)
	- [Soccer World's Cup 12/7/98] France/Brasil : 3-0 :-)

	Version 0.7.2 :
	- oops, looks like I forgot to put the log line...
	- Intel release

	Version 0.7.1 :
	- did I really output a version numbered 0.7 before ? I can't remember :(
	- Integrate some useful stuff from Bernie Estavillo (bernie@metrowerks.com)
	1. Unique file names so the new archive will not overwrite an existing archive.
	2. Prepend the date to the file name when the Shift key is held down, so an
	   archive of a directory called test will end up with a name like
	   1997-08-13.test.zip. I'm not sure how useful this will be to others.
	- I also modified Bernie's nice code so that the date can be appended in the
	  form of zipname_ddmmyyyy.zip if you hold down the Control key (don't blame me, it's
	  more natural for me to see dates in the "other" way, i.e. the French one :))
	  If you hold both Control and Shift, you get the append date.

	Version 0.6 :
	- now gracefully handles files containing spaces, quotes, double
	  quotes and others characters interpreted by the shell. Thx to GL !

	For example, the file :
	"./#zi|;p ~m^ &$e_06 _P&R.t)xt.z ip.z_ @ip.z\}ip.zip&(-_.zip)=~#{[|`\^@]}^$¨$*%!;,?.§<stuff>dyn"
	can now be correctly zipped (I've tried ;))

	Version 0.0 :
	- Based on code from TermHire.cpp, v0.97, release date 24/05/97 by
	  Pierre BRUA (brua@dess-info.u-strasbg.fr
	
	---
	USE AT YOUR OWN RISK !

*************************************************************************/

#include <Directory.h>
#include <Window.h>
#include <StringView.h>
#include <File.h>
#include <Alert.h>
#include <Message.h>
#include <Errors.h>
#include <Path.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

// this tell the compiler that process_refs will be called from
// outside (by the Tracker)

#include <TrackerAddOn.h>

//------------------------------------------------------------------------
// some defines...

#pragma once

// use this define if you want little windows to explain you what's going on
#define VERBOSE

// use this define if you want printf() in the terminal (run a terminal,
// use Olivier's TManager to quit the Tracker, type /boot/system/Tracker & in
// the terminal and you get the stdout here. Too cool !
#undef DPRINTF
//#define DPRINTF

// the name of the add_on
#define ADDON_NAME "ZipMe"
#define ADDON_VERS "0.7.4"

// the default MIME type that Zip files should get
#define ZIP_FILE_MIMETYPE	"application/x-zip"

// change behavior as you want it to. It's a bad idea to remove -r for directory handling !
#define K_ZIP_OPTIONS "-ryq"

#ifndef max
#define max(a,b) ((a)>(b) ? (a) : (b))
#endif

//------------------------------------------------------------------------
// internal state. I use a struct rather than global variables.

struct SState
{
	BWindow *win;
	BTextControl *leaf, *dir, *lasterr, *method, *number;
	BStatusBar *bar;
	int32 errorCount;
	int32 totalCount;
	int32 baseCount;
	bool appendDate;
	bool prependDate;
};


//------------------------------------------------------------------------
// local prototypes 
void error(const char *message);
void createWindow(SState &state);
void closeWindow(SState &state);
void setFileInfo(SState &state, const char *leaf, const char *dir);
void setError(SState &state, const char *error);
void setMethod(SState &state, const char *text);
void setNumber(SState &state, int32 count);
char* uniqueName(SState &state, BPath path);
void installOurself(void);

void packFileOrDir(SState &state, entry_ref &ref, bool isdir);


/* Code to help a little the fellow user double-clicking the
   add-on icon directly.
*/
int main(int, char **)
{
	new BApplication("application/x-vnd.ralf-vanilla");
	BAlert *box = new BAlert(ADDON_NAME,
													 "Thanks for using " ADDON_NAME " " ADDON_VERS " !\n\n"
													 "Please move the " ADDON_NAME " executable into\n"
													 "/boot/home/config/add-ons/Tracker/\n"
													 "and try again !",
													 "Don't Touch Me !", "Install Myself !", NULL,
													 B_WIDTH_FROM_WIDEST);
	if (box->Go() == 1) installOurself();

	delete be_app;
	return 0;
}


//***********************
void installOurself(void)
//***********************
/*
	Copy *this* binary file into /boot/home/config/add-ons/Tracker/,
	making sure the directories exists and can be created.
*/
{
char appName[B_FILE_NAME_LENGTH];
app_info info_app;
BEntry file_app;
BPath	idealPath;
BDirectory addonsDir;
BDirectory trackerDir;
status_t result;

	be_app->GetAppInfo(&info_app);
	result=file_app.SetTo(&info_app.ref);
	if (result != B_NO_ERROR) goto bail_out_error;

	// The ideal directory is B_USER_ADDONS_DIRECTORY+"./Tracker"
	// create it if doesn't exists

	result = find_directory(B_USER_ADDONS_DIRECTORY, &idealPath, true);
	if (result != B_NO_ERROR && result != B_FILE_EXISTS) goto bail_out_error;

	if (addonsDir.SetTo(idealPath.Path()) != B_NO_ERROR) goto bail_out_error;
	result = addonsDir.CreateDirectory("Tracker", &trackerDir);
	if (result == B_FILE_EXISTS) result = trackerDir.SetTo(&addonsDir, "Tracker");
	if (result != B_NO_ERROR) goto bail_out_error;
	if (trackerDir.InitCheck() != B_NO_ERROR) goto bail_out_error;

	// move executable
	result = file_app.MoveTo(&trackerDir, NULL, true);
	if (result != B_NO_ERROR && result != B_FILE_EXISTS) goto bail_out_error;

	return;

bail_out_error:
	error("Couldn't toss the directories and the file as expected.");
}


//*****************************
void error(const char *message)
//*****************************
/*
	Displays a message in a dialog box (yes, believe it!)
*/
{
	BAlert *alert = new BAlert(ADDON_NAME " Error :\n", message, "OK");
	alert->Go();
}


//********************************************************************************
char* uniqueName(SState &state, BPath path)
//********************************************************************************
/*
	Based on idea from: Bernie Estavillo <bernie@austin.metrowerks.com>
	Date: Wed, 13 Aug 1997 11:00:58 -0500
*/
{
    char dateString[16] = "";
    char baseName[B_FILE_NAME_LENGTH * 2];

    // append the current date to the beginning of the file name
    if (state.appendDate || state.prependDate)
    {
        time_t          now;
        struct tm       *date;
        now = time(NULL);
        date = localtime(&now);
        if (state.appendDate) strftime(dateString, 16, "_%d%m%Y", date);
        else strftime(dateString, 16, "%Y-%m-%d.", date);
    }
	if (state.appendDate) sprintf(baseName, "%s%s", path.Leaf(), dateString);
	else if (state.prependDate) sprintf(baseName, "%s%s", dateString, path.Leaf());
    else strcpy(baseName, path.Leaf());

    // ensure that the final file name is unique
    BPath                tempPath(path);
    path.GetParent(&tempPath);
    BDirectory           targetDir(tempPath.Path());
    char                 zipName[B_FILE_NAME_LENGTH * 2];
    sprintf(zipName, "%s.zip", baseName);

    int32 count = 0;
    while(targetDir.Contains(zipName))
    {
        sprintf(zipName, "%s.%d.zip", baseName, ++count);
    }

    char *outName = new char[B_FILE_NAME_LENGTH * 2 + 128];
    if (outName) strcpy(outName, zipName);

    return outName;
}


//********************************************************************************
void packFileOrDir(SState &state, entry_ref &ref, bool isdir)
//********************************************************************************
/*
	Version update:
	- there was previously two functions, packDirectory() and packFile() but the code
	  was so similar that I packed them into only one. From a Zip point-of-view, packing
	  a dir or a file changes nothing, only the setMethod() comment changes ;)
	  (at first, I thougth to add some tricky code inside packDir to display names of
	  packed files on the fly. Finally I lazily forgot the stuff.)
	- this now calls uniqueName(), idea from Bernie, make zip names uniques and add a date
	  if needed.
*/
{
BEntry entry(&ref);
BEntry parent;
BPath path;
BPath parentPath;
char buf[B_FILE_NAME_LENGTH*2+128];

	entry.GetParent(&parent);
	entry.GetPath(&path);
	parent.GetPath(&parentPath);

	setFileInfo(state, path.Leaf(), parentPath.Path());

	// chdir might not be necessary -- unless I do recursive handling
	chdir(parentPath.Path());

	// compute unique zip name
	char *zipname = uniqueName(state, path);
	if (!zipname)
	{
		error("Memory allocationg error in " ADDON_NAME ".");
		return;
	}

	// modification proposed by Guillaume Lefevre - 06 Aug 97
	// idea is to put quotes or double quotes around the name...
	if(strchr(zipname,'\'') != NULL)
		sprintf(buf, "zip %s \"%s\" \"%s\"", K_ZIP_OPTIONS, zipname, path.Leaf());
	else
		sprintf(buf, "zip %s '%s' '%s'", K_ZIP_OPTIONS, zipname, path.Leaf());
	
	if (isdir) setMethod(state, "zip directory recursively");
	else setMethod(state, "zip single file");

	// issue the real action...
	// error checking by Peter Hinely <phinely@hawaii.edu>, 14 Jul 1998
  if(system(buf) != 0)
  {
		BAlert *alert = new BAlert(ADDON_NAME ADDON_VERS,
																ADDON_NAME ADDON_VERS " : " 
																"Error during creation of zip archive.",
																"OK", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT);
		alert->Go();
  }

	// make sure that the MIME type of the file is "application/x-zip".
	// R3.0 zip binary creates a wrong "application/zip" type.
	// Note : this looks like the Rigth Thing [tm] to do but unfortunately
	// here Expander won't recognize the correct signature and will prefer the
	// bad one. Time for a [KEY BE]... :-)
	if(0)
	{
		BDirectory dir_parent;
		if (entry.GetParent(&dir_parent) >= B_NO_ERROR)
		{
			BNode zipnode;
			BNodeInfo zipinfo;
			if (zipnode.SetTo(&dir_parent, zipname) >= B_NO_ERROR
					&& zipinfo.SetTo(&zipnode) >= B_NO_ERROR)
				zipinfo.SetType(ZIP_FILE_MIMETYPE);
		}
	}

	// release the memory allocated by uniqueName() in zipname
	// memory leak detected by Peter Hinely <phinely@hawaii.edu>, 14 Jul 1998
	delete zipname;
}


//********************************************************
void process_refs(entry_ref dir_ref, BMessage* msg, void*)
//********************************************************
/*
	this is the function the Tracker call when the add-on is selected
	dir_ref is the folder you have launched it from
	msg is a BMessage containing the references to BEntry's(Files/Folders)
	and for "void *", I don't know yet...
*/
{
int32 count;	// number of references stored in the BMessage
ulong type;		// type of the BMessage
SState state;
BPath path;

	// be clean and please respect the OS.h header allegory.
	if (!is_computer_on())
	{
		error("Please power-on your computer before running " ADDON_NAME);
		return;
	}

	// might check is_computer_on_fire() to be clean, too

	state.win = NULL;
	state.errorCount = 0;
	state.totalCount = 0;
	state.baseCount = 0;
    state.prependDate = ((modifiers() & B_SHIFT_KEY) != 0);
    state.appendDate = ((modifiers() & B_CONTROL_KEY) != 0);

	// this creates an informative window about the completion of the add-on
	#ifdef VERBOSE
		createWindow(state);
	#endif

	// first, we consider the case when the add-on has been selected
	// from the file menu of the browser window or by right-clicking
	// in the background of the window
	// i.e : in this case, there is no refs on the BMessage
	msg->GetInfo("refs", &type, &count);
	if (count == 0)
	{
		error("Please select a file or a directory to compress !");
	}
	else
	{
		BDirectory dir;
		BEntry entry;

		// if there are datas on the BMessage, but not from the
		// directory type, exit...
		if (dir.SetTo(&dir_ref) != B_NO_ERROR) 
		{
			error("You must launch " ADDON_NAME " from folders.");
			return;
		}
		if(dir.GetEntry(&entry) != B_NO_ERROR)
		{
			error("Sorry, I cannot access the folder while launching " ADDON_NAME);
			return;
		}

		state.totalCount = count;

		#ifdef VERBOSE
		if (state.win && state.bar)
		{
			state.win->Lock();
			state.bar->SetMaxValue(count);
			state.win->Unlock();
		}
		#endif

		try
		{
			for(long i=0; i<count; i++)
			{
				// pack one file or one directory...
				entry_ref ref;
				BDirectory dir;
	
				setNumber(state, i);
				if (msg->FindRef("refs", i, &ref) >= B_NO_ERROR)
				{
					// if the ref is a folder, expand the directory content
					if(dir.SetTo(&ref) >= B_NO_ERROR) packFileOrDir(state, ref, true);
					else
					{
						BFile file;
						if(file.SetTo(&ref, B_READ_ONLY) >= B_NO_ERROR) packFileOrDir(state, ref, false);
					}
				}
			}
		}
		catch(...)
		{
			setError(state, "Bad luck !\nException throw !\nProbably IO or File Exception.");
		}
	}
	
	// exit...

	if (state.errorCount > 0)
	{
		char s[256];
		sprintf(s, ADDON_NAME " v." ADDON_VERS " :\n%d error%s detected.",
			state.errorCount, (state.errorCount > 1 ? "s" : ""));
		#ifdef VERBOSE
			if (state.win) state.win->Activate();
		#endif
		error(s);
	}

	#ifdef VERBOSE
		closeWindow(state);
	#endif
}


//******************************
void createWindow(SState &state)
//******************************
{
	BRect area(0,0, 280,80);

	BScreen screen;		// -- DR9 PR
	BRect r=screen.Frame();
	area.OffsetBy(r.Width()/2-125, r.Height()/2-80);

	state.win = new BWindow(area, ADDON_NAME " Status", B_TITLED_WINDOW,
							B_NOT_ZOOMABLE | B_NOT_V_RESIZABLE | B_NOT_CLOSABLE);
	if (!state.win) return;

	BRect bounds = state.win->Bounds();
	BView *base = new BView(bounds, B_EMPTY_STRING, B_FOLLOW_ALL, B_WILL_DRAW);
	base->SetViewColor(230, 230, 230);
	if (!base) goto error;	// I could have used an exception...
	state.win->AddChild(base);

	BRect rect;
	rect.left		= bounds.left  + 10.0;
	rect.top		= bounds.top   + 10.0;
	rect.right	= bounds.right - 10.0;
	rect.bottom	= rect.top+15;

	uint32 rmask = B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP;
	BStringView *str= new BStringView(rect, B_EMPTY_STRING,
						ADDON_NAME " v." ADDON_VERS ", by R'alf", rmask, B_WILL_DRAW);
	if (!str) goto error;	
	base->AddChild(str);

	rect = str->Frame();
	rect.OffsetBy(0.0, rect.Height()+ 20.0);
	state.leaf		= new BTextControl(rect, B_EMPTY_STRING, "File name", "", NULL, rmask);
	if (!state.leaf) goto error;
	base->AddChild(state.leaf);

	rect = state.leaf->Frame();
	rect.OffsetBy(0.0, rect.Height()+ 10.0);
	state.dir			= new BTextControl(rect, B_EMPTY_STRING, "Directory", "", NULL, rmask);
	if (!state.dir) goto error;
	base->AddChild(state.dir);

	rect.OffsetBy(0.0, rect.Height()+ 10.0);
	state.lasterr	= new BTextControl(rect, B_EMPTY_STRING, "Last Error", "", NULL, rmask);
	if (!state.lasterr) goto error;
	base->AddChild(state.lasterr);

	rect.OffsetBy(0.0, rect.Height()+ 10.0);
	state.method	= new BTextControl(rect, B_EMPTY_STRING, "Running", "", NULL, rmask);
	if (!state.method) goto error;
	base->AddChild(state.method);

	state.number = NULL;

	rect.OffsetBy(0.0, rect.Height()+10.0);
	state.bar = new BStatusBar(rect, B_EMPTY_STRING, "File count", NULL /*, B_WILL_DRAW, rmask*/);
	if (!state.bar) goto error;
	state.bar->SetFlags(B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE);
	state.bar->SetResizingMode(rmask);
	base->AddChild(state.bar);
	rect = state.bar->Frame();

	long n=20;
	n = max(state.leaf   ->StringWidth("File name" ), n);
	n = max(state.dir    ->StringWidth("Directory" ), n);
	n = max(state.lasterr->StringWidth("Last Error"), n);
	n = max(state.method ->StringWidth("Running"   ), n);
	state.leaf   ->SetDivider(n + 2.0);
	state.dir    ->SetDivider(n + 2.0);
	state.lasterr->SetDivider(n + 2.0);
	state.method ->SetDivider(n + 2.0);

	state.leaf   ->SetFlags(B_WILL_DRAW);
	state.dir    ->SetFlags(B_WILL_DRAW);
	state.lasterr->SetFlags(B_WILL_DRAW);
	state.method ->SetFlags(B_WILL_DRAW);

	state.win->ResizeTo(bounds.Width(), rect.bottom+10.0);
	state.win->SetSizeLimits(bounds.Width(), r.Width()-20.0, rect.bottom+10.0, rect.bottom+10.0);
	state.win->Show();
	return;

error:
	state.win = NULL;

} // end of createWindow


//******************************
void 	closeWindow(SState &state)
//******************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;

	state.win->Lock();
	state.win->Quit();
	int32 val;
	wait_for_thread(state.win->Thread(), &val);

} // end of closeWindow



//****************************************************************
void 	setFileInfo(SState &state, const char *leaf, const char *dir)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	if (leaf && state.leaf) state.leaf->SetText(leaf);
	if (dir  && state.dir ) state.dir ->SetText(dir );
	state.win->Unlock();
} // end of setFileInfo


//****************************************************************
void 	setError(SState &state, const char *error)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	if (error && state.lasterr) state.lasterr->SetText(error);
	state.win->Unlock();
} // end of setError


//****************************************************************
void 	setMethod(SState &state, const char *text)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	if (text && state.method) state.method->SetText(text);
	state.win->Unlock();
} // end of setMethod


//****************************************************************
void 	setNumber(SState &state, int32 count)
//****************************************************************
{
char s[256];
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	sprintf(s, "%d of %d", state.baseCount+count, state.totalCount);
	state.win->Lock();
	if (state.bar) state.bar->Update(1.0, NULL, s);
	state.win->Unlock();
} // end of setNumber


// eoc

